from django.urls import include, path

urlpatterns = [
    # Make this distinct from typical production values, to ensure it works dynamically
    path('rgd_test/', include('rgd.urls')),
    path('rgd_geometry_test/', include('rgd_geometry.urls')),
]



from rest_framework.generics import RetrieveAPIView
from rgd.rest.get import _PermissionMixin
from rgd_geometry import models, serializers


class GetGeometry(RetrieveAPIView, _PermissionMixin):
    serializer_class = serializers.GeometrySerializer
    lookup_field = 'pk'
    queryset = models.Geometry.objects.all()


class GetGeometryData(RetrieveAPIView, _PermissionMixin):
    serializer_class = serializers.GeometryDataSerializer
    lookup_field = 'pk'
    queryset = models.Geometry.objects.all()



import factory
import factory.django
from rgd_geometry import models
from rgd_testing_utils.factories import ChecksumFileFactory


class GeometryArchiveFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.GeometryArchive

    file = factory.SubFactory(ChecksumFileFactory)



import pytest
from rgd.datastore import datastore
from rgd_geometry import models
from rgd_geometry.tasks.etl import read_geometry_archive

from . import factories


@pytest.mark.parametrize(
    'sample_file',
    ['Streams.zip', 'dlwatersan.zip'],
)
@pytest.mark.django_db(transaction=True)
def test_geometry_etl(sample_file):
    geom_archive = factories.GeometryArchiveFactory(
        file__file__filename=sample_file,
        file__file__from_path=datastore.fetch(sample_file),
    )
    entry = models.Geometry.objects.filter(geometry_archive=geom_archive).first()
    assert entry.data is not None
    # Testing that we can repopulate a geometry archive again
    read_geometry_archive(geom_archive.id)
    # test the field file validator
    models.validate_archive(geom_archive.file.file)



from django.contrib.gis.db import models
from django.core.exceptions import ValidationError
from django_extensions.db.models import TimeStampedModel
import magic
from rgd.models import ChecksumFile, SpatialEntry
from rgd.models.constants import DB_SRID
from rgd.models.mixins import DetailViewMixin, PermissionPathMixin, TaskEventMixin
from rgd_geometry.tasks import jobs


def validate_archive(field_file):
    """Validate file is a zip or tar archive."""
    acceptable = ['application/zip', 'application/gzip']

    mimetype = magic.from_buffer(field_file.read(16384), mime=True)

    if mimetype not in acceptable:
        raise ValidationError('Unsupported file archive.')


class GeometryArchive(TimeStampedModel, TaskEventMixin, PermissionPathMixin):
    """Container for ``zip`` archives of a shapefile.

    When this model is created, it loads data from an archive into
    a single ``Geometry`` that is then associated with this entry.
    """

    task_funcs = (jobs.task_read_geometry_archive,)
    file = models.ForeignKey(ChecksumFile, on_delete=models.CASCADE, related_name='+')

    def save(self, *args, **kwargs):
        super(GeometryArchive, self).save(*args, **kwargs)

    def archive_data_link(self):
        return self.file.data_link()

    archive_data_link.allow_tags = True

    permissions_paths = [('file', ChecksumFile)]


class Geometry(TimeStampedModel, SpatialEntry, PermissionPathMixin, DetailViewMixin):
    """A holder for geometry vector data."""

    name = models.CharField(max_length=1000, blank=True)
    description = models.TextField(null=True, blank=True)

    data = models.GeometryCollectionField(srid=DB_SRID)  # Can be one or many features
    # The actual collection is iterable so access is super easy

    # Can be null if not generated from uploaded ZIP file but something else
    geometry_archive = models.OneToOneField(GeometryArchive, null=True, on_delete=models.CASCADE)

    permissions_paths = [('geometry_archive', GeometryArchive)]
    detail_view_name = 'geometry-entry-detail'



from rgd_testing_utils.settings import *  # noqa

INSTALLED_APPS += [  # noqa
    'rgd_geometry',
]

ROOT_URLCONF = 'test_project.urls'

CELERY_TASK_ALWAYS_EAGER = True
CELERY_TASK_EAGER_PROPAGATES = True



import logging

from rgd.management.commands._data_helper import SynchronousTasksCommand, _get_or_create_file_model
from rgd_geometry import models

logger = logging.getLogger(__name__)


SUCCESS_MSG = 'Finished loading all demo data.'

# Names of files in the datastore
SHAPE_FILES = [
    'Streams.zip',
    'Watershedt.zip',
    'MuniBounds.zip',
    'lm_cnty.zip',
    'dlwatersan.zip',
    'dlschool.zip',
    'dlpark.zip',
    'dlmetro.zip',
    'dllibrary.zip',
    'dlhospital.zip',
    'dlfire.zip',
    'Solid_Mineral_lease_1.zip',
    'AG_lease.zip',
]


def load_shape_files(shape_files):
    ids = []
    for shpfile in shape_files:
        entry = _get_or_create_file_model(models.GeometryArchive, shpfile)
        ids.append(entry.geometrymeta.pk)
    return ids


class Command(SynchronousTasksCommand):
    help = 'Populate database with demo data.'

    def handle(self, *args, **options):
        self.set_synchronous()
        # Run the command
        load_shape_files(SHAPE_FILES)

        self.stdout.write(self.style.SUCCESS(SUCCESS_MSG))
        self.reset_celery()



from django.urls import path

from . import models, rest, views

urlpatterns = [
    # Pages
    path(
        'rgd_geometry/<int:pk>/',
        views.GeometryDetailView.as_view(),
        name=models.Geometry.detail_view_name,
    ),
    #############
    # Other
    path(
        'api/rgd_geometry/<int:pk>',
        rest.get.GetGeometry.as_view(),
        name='geometry-entry',
    ),
    path(
        'api/rgd_geometry/<int:pk>/data',
        rest.get.GetGeometryData.as_view(),
        name='geometry-entry-data',
    ),
]



from django.apps import AppConfig


class RGDGeometryConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'rgd_geometry'

    def ready(self):
        import rgd_geometry.signals  # noqa: F401



import json

from rgd import utility
from rgd.serializers import SpatialEntrySerializer

from . import models


class GeometrySerializer(SpatialEntrySerializer):
    class Meta:
        model = models.Geometry
        exclude = ['data', 'footprint', 'outline']


class GeometryDataSerializer(GeometrySerializer):
    def to_representation(self, value):
        ret = super().to_representation(value)
        ret['data'] = json.loads(value.data.geojson)
        return ret

    class Meta:
        model = models.Geometry
        fields = '__all__'


utility.make_serializers(globals(), models)



"""Helper methods for creating a geometry entries from uploaded files."""
from glob import glob
import os
import tempfile
import zipfile

from celery.utils.log import get_task_logger
from django.conf import settings
from django.contrib.gis.gdal import SpatialReference
from django.contrib.gis.geos import GeometryCollection, GEOSGeometry, Polygon
from django.core.exceptions import ValidationError
import fiona
from rgd.models.transform import transform_geometry
from rgd.utility import get_or_create_no_commit
from rgd_geometry.models import Geometry, GeometryArchive
from shapely.geometry import shape
from shapely.wkb import dumps

logger = get_task_logger(__name__)


def read_geometry_archive(archive_pk):
    """Read an archive of a single shapefile (and associated) files.

    This will load zipped archives of shape files and create entries
    for a single shapefile (basename of files).

    A single shapefile will consist of a collection of one or many features
    of varying types. We produce a single ``GeometryCollection`` of those
    data. Hence, we associate a single shapefile with a single
    ``GeometryCollection``.

    We may need to do more checks/validation to make sure the user only
    added a single shape file or provide a more explicit upload interface
    where they upload the ``shp``, ``dbf``, etc. files individually and
    we assert that they match.

    """
    archive = GeometryArchive.objects.get(pk=archive_pk)

    # TODO: add a setting like this:
    workdir = getattr(settings, 'GEODATA_WORKDIR', None)
    with tempfile.TemporaryDirectory(dir=workdir) as tmpdir:
        with archive.file.yield_local_path() as archive_path:
            logger.info(f'The geometry archive: {archive_path}')

            # Unzip the contents to the working dir
            with zipfile.ZipFile(archive_path, 'r') as zip_ref:
                zip_ref.extractall(tmpdir)

        msg = 'There must be one and only one shapefile in the archive. Found ({})'
        shape_files = glob(os.path.join(tmpdir, '*.shp'))
        if len(shape_files) > 1:
            raise ValidationError(msg.format(len(shape_files)))
        elif len(shape_files) == 0:
            shape_files = glob(os.path.join(tmpdir, '**/*.shp'))
            if len(shape_files) != 1:
                raise ValidationError(msg.format(len(shape_files)))
        shape_file = shape_files[0]

        # load each shapefile using fiona
        with fiona.open(shape_file) as shapes:
            geometry_entry, created = get_or_create_no_commit(
                Geometry, defaults=dict(name=archive.file.name), geometry_archive=archive
            )

            shapes.meta  # TODO: dump this JSON into the model entry

            crs_wkt = shapes.meta['crs_wkt']
            logger.info(f'Geometry crs_wkt: {crs_wkt}')
            spatial_ref = SpatialReference(crs_wkt)
            logger.info(f'Geometry SRID: {spatial_ref.srid}')

            collection = []
            for item in shapes:
                geom = shape(item['geometry'])  # not optimal?
                # TODO: check this
                collection.append(
                    transform_geometry(
                        GEOSGeometry(
                            memoryview(dumps(geom, srid=spatial_ref.srid)), srid=spatial_ref.srid
                        ),
                        crs_wkt,
                    )
                )

    geometry_entry.data = GeometryCollection(*collection)
    geometry_entry.footprint = geometry_entry.data.convex_hull
    bounds = geometry_entry.footprint.extent  # (xmin, ymin, xmax, ymax)
    coords = [
        (bounds[0], bounds[3]),  # (xmin, ymax)
        (bounds[2], bounds[3]),  # (xmax, ymax)
        (bounds[2], bounds[1]),  # (xmax, ymin)
        (bounds[0], bounds[1]),  # (xmin, ymin)
        (bounds[0], bounds[3]),  # Close the loop
    ]
    geometry_entry.outline = Polygon(coords)

    geometry_entry.save()

    return True



from django.contrib import admin
from django.contrib.gis.admin import OSMGeoAdmin
from rgd.admin.mixins import (
    MODIFIABLE_FILTERS,
    SPATIAL_ENTRY_FILTERS,
    TASK_EVENT_FILTERS,
    TASK_EVENT_READONLY,
    GeoAdminInline,
    _FileGetNameMixin,
    reprocess,
)
from rgd_geometry.models import Geometry, GeometryArchive


class GeometryInline(GeoAdminInline):
    model = Geometry
    fk_name = 'geometry_archive'
    list_display = (
        'pk',
        'name',
        'geometry_archive',
        'modified',
        'created',
    )
    list_filter = MODIFIABLE_FILTERS + SPATIAL_ENTRY_FILTERS
    readonly_fields = (
        'modified',
        'created',
        'geometry_archive',
    )
    modifiable = False  # To still show the footprint and outline


@admin.register(GeometryArchive)
class GeometryArchiveAdmin(OSMGeoAdmin, _FileGetNameMixin):
    list_display = (
        'pk',
        'get_name',
        'status',
        'modified',
        'created',
        'archive_data_link',
    )
    readonly_fields = (
        'modified',
        'created',
    ) + TASK_EVENT_READONLY
    inlines = (GeometryInline,)
    actions = (reprocess,)
    list_filter = MODIFIABLE_FILTERS + TASK_EVENT_FILTERS



from rgd.views import _SpatialDetailView

from . import models


class GeometryDetailView(_SpatialDetailView):
    model = models.Geometry

    def _get_extent(self):
        extent = super()._get_extent()
        extent['data'] = self.object.data.json
        return extent



from celery import shared_task
from rgd.tasks import helpers


@shared_task(time_limit=86400)
def task_read_geometry_archive(archive_pk):
    from rgd_geometry.models import GeometryArchive

    from .etl import read_geometry_archive

    archive = GeometryArchive.objects.get(pk=archive_pk)
    helpers._run_with_failure_reason(archive, read_geometry_archive, archive_pk)



from io import open as io_open
import os
from pathlib import Path

from setuptools import find_packages, setup

readme_file = Path(__file__).parent.parent / 'README.md'
if readme_file.exists():
    with readme_file.open() as f:
        long_description = f.read()
else:
    # When this is first installed in development Docker, README.md is not available
    long_description = ''

__version__ = None
filepath = os.path.dirname(__file__)
version_file = os.path.join(filepath, '..', 'version.py')
with io_open(version_file, mode='r') as fd:
    exec(fd.read())

setup(
    name='django-rgd-geometry',
    version=__version__,
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    author='Kitware, Inc.',
    author_email='rgd@kitware.com',
    url='https://github.com/ResonantGeoData/ResonantGeoData',
    keywords='',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django :: 3.0',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python',
    ],
    python_requires='>=3.8',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'django-rgd',
    ],
    extras_require={
        'worker': [
            'shapely',
            'fiona',
        ],
    },
)
